/*  User changes  ----------------------------------------------------------

	The user may change the following to take account of their own
	specific circumstances:

	#defines:

		RECORD_BUFFER_SIZE
		NO_OF_FIELDS

	Source code:

		Line 41	Union may change to reflect different field types
		Line 114	May change according to user fields
		Line 134	May omit

	The only other change which I should make is to allocate the
	Fields[] dynamically using p_alloc() rather than allocating
	it all at once which is very antisocial.
    ------------------------------------------------------------------------  */

#include <epoc.h>
#include <p_std.h>
#include <p_file.h>
#include <p_dbf.h>

#define					RECORD_BUFFER_SIZE	4096
#define					MAX_FIELDS				32
#define					NO_OF_FIELDS			8
#define					MAX_FIELD_LENGTH		128


/*  ------------------------------------------------------------------------
	 The following union is used to store the data extracted from the  ------
	 individual fields. It's a union because fields can be, with this  ------ 
	 code, a string, a word or a double. These are the field types used  ----
	 if we wish to remain OPL compatible. If would be simple to alter  ------
	 the code to take account of, for example, unsigned fields.  ------------
	 ------------------------------------------------------------------------  */

typedef union
	{
	TEXT						stringfield[MAX_FIELD_LENGTH];
	INT						integerfield;
	DOUBLE					floatfield;
	LONG						longfield;
	} DBFFIELD;


LOCAL_D INT					ret;                  /*  General purpose return var  */
LOCAL_D INT					i;                   /*  General purpose counter var  */

LOCAL_D UINT				RecordCount;     /*  Total number of records in file  */

LOCAL_D DbfOpenArgs		FileArgs;        /*  Used for file handle, mode, etc  */
LOCAL_D DbfHeader  		Header;                              /*  File header  */
LOCAL_D VOID			*	Fcb	=	NULL;                       /*  File handle  */
LOCAL_D UINT				FileState;                          /*  Index option  */
LOCAL_D UBYTE				RecordBuffer[RECORD_BUFFER_SIZE];     /*  The buffer  */

/*  ------------------------------------------------------------------------
    The data extracted from the record buffer is stored in the following  --
    array of DBFFIELDs. The programmer can imagine these simply as fields. -
    So, in the OPL way that they picture A.a$ being a string field, they  --
    can picture Fields[0].stringfield being a string field.  ---------------
    ------------------------------------------------------------------------  */

LOCAL_D DBFFIELD			Fields[NO_OF_FIELDS];


/*  ------------------------------------------------------------------------    
    General purpose error check. If val is an error,  ----------------------
    puts up a notifier and then just exits. If val  ------------------------
    is good, just returns val. ---------------------------------------------
    ------------------------------------------------------------------------  */

LOCAL_C INT Check(INT val, TEXT *msg)
    {
    TEXT						buf[50];

    if (val >= 0)
        return val;
    p_atos(&buf[0],"Error : %s",msg);
    p_notifyerr(val,&buf[0],0,0,0);
    p_exit(0);
    }


/*  ------------------------------------------------------------------------
    Opens the Dbf file. The only interesting point here  -------------------
    is the fact that the file is opened with the DbfStateStart  ------------
    argument. This is good practise and allows the application  ------------
    to remain responsive to the user (perhaps to allow a cancel  -----------
    when opening a particularly large file).  ------------------------------
    ------------------------------------------------------------------------  */

LOCAL_C VOID OpenFile(TEXT *filename)
	{
	p_scpy((TEXT *)Header.fileType,"OPLDatabaseFile");
	Header.createVersion	=	DbfVersion();
	Header.dataStart		=	DbfHeaderLength;
	Header.needVersion		=	DbfVersion();
	Header.firHeader		=	((DbfRecordTypeFir << 12) | NO_OF_FIELDS);

/*  ------------------------------------------------------------------------
    The following section may vary in length. This  ------------------------
    is where we define the field types. It is entirely possible  -----------
    to simply define 1 field at this stage. Read page 183  -----------------
    of the PLIB manual for further details. At this time, make  ------------
    sure that you define the fields that you are really going  -------------
    to use as, for the time being, the GetFields() function  ---------------
    uses this data as a reference.  ----------------------------------------
    ------------------------------------------------------------------------  */
	
	Header.fir[0]			=	DbfFieldTypeString;
	Header.fir[1]			=	DbfFieldTypeString;
	Header.fir[2]			=	DbfFieldTypeString;
	Header.fir[3]			=	DbfFieldTypeWord;
	Header.fir[4]			=	DbfFieldTypeDouble;
	Header.fir[5]			=	DbfFieldTypeWord;
	Header.fir[6]			=	DbfFieldTypeLong;
	Header.fir[7]			=	DbfFieldTypeString;

	FileState				=	DbfStateStart;
	FileArgs.pFcb			=	&Fcb;
	FileArgs.fName			=	filename;
	FileArgs.mode			=	P_FOPEN;           /*  Open an existing file  */
	FileArgs.pHead			=	&Header;

	do
		{
		ret	=	DbfQuickOpen(	(INT *)&FileState,
								&FileArgs,
								(unsigned char *)&RecordBuffer[0],
								RECORD_BUFFER_SIZE,
								1);

		} while((ret >= 0) && (FileState != DbfStateStart));

	RecordCount	=	DbfCount(Fcb);         /*  Just a simple debugging check  */
	}
	
	
LOCAL_C VOID GetFields(VOID)
	{
	UBYTE					*	pBuffer;            /*  Pointer to the record buffer  */
	WORD						RecordEnd;          /*  Address of end of the record  */
	WORD						RecordLength;               /*  Length of the record  */
	INT						CopyLen;                 /*  Number of bytes to copy  */
	INT						FIRFlag = FALSE;    /*  Do records differ from FIR ?  */
	

	pBuffer = &RecordBuffer[0];                  /*  Point to start of buffer  */
	RecordLength = *(WORD *)pBuffer;               /*  Get the record length   */
	RecordLength = RecordLength & 0x0fff;     /*  We don't want the MS nibble  */
	RecordEnd = (WORD)pBuffer + RecordLength;
                                           /*  Note address of end of record  */
															 
	pBuffer += 2;                                   /*  Skip the header bytes  */
	
	i = 0;                                /*  Initialise the counter which we  */
	                                          /*  will use to increment field  */
	do
		{
		if (FIRFlag)         /*  If flag has been set, records differ from FIR  */
			{              /*  we assume the fields are strings (PLIB page 183)  */
			CopyLen = *pBuffer;                 /*  Get the string length (LCB)  */
			pBuffer++;                                         /*  Skip the LCB  */
			*p_bcpy(&Fields[i].stringfield[0], pBuffer, CopyLen) = 0;
                           /*  Extract the data from this field to the array  */
			pBuffer += CopyLen;                 /*  Skip to start of next field  */
			}
		else if (Header.fir[i] == DbfFieldTypeString)     /*  If it's a string  */
			{
			CopyLen = *pBuffer;                 /*  Get the string length (LCB)  */
			pBuffer++;                                         /*  Skip the LCB  */
			*p_bcpy(&Fields[i].stringfield[0], pBuffer, CopyLen) = 0;
                           /*  Extract the data from this field to the array  */
			pBuffer += CopyLen;                 /*  Skip to start of next field  */
			}
		else if(Header.fir[i] == DbfFieldTypeWord)          /*  If it's a word  */
			{
			Fields[i].integerfield = *(INT *)pBuffer;
                           /*  Extract the data from this field to the array  */
			pBuffer += sizeof(INT);             /*  Skip to start of next field  */
			}
		else if(Header.fir[i] == DbfFieldTypeLong)          /*  If it's a long  */
			{
			Fields[i].longfield = *(LONG *)pBuffer;
                           /*  Extract the data from this field to the array  */
			pBuffer += sizeof(LONG);            /*  Skip to start of next field  */
			}
		else if(Header.fir[i] == DbfFieldTypeDouble)       /*  If it's a float  */
			{
			Fields[i].floatfield = *(DOUBLE *)pBuffer;
                           /*  Extract the data from this field to the array  */
			pBuffer += sizeof(DOUBLE);          /*  Skip to start of next field  */
			}
/*		i++;                                           Increment the counter  */
		if (++i > NO_OF_FIELDS)                         /*  Has the FIR lied ?  */
			FIRFlag = TRUE;                           /*  Note a trailing field  */
		} while ((WORD)pBuffer < RecordEnd);           /*  Are we at the end ?  */
	}
	


LOCAL_C VOID CleanUpAndQuit(VOID)
	{
	DbfClose(Fcb);                                         /*  Close the file  */
	p_exit(0);                                               /*  Exit cleanly  */
	}
	


GLDEF_C VOID main(VOID)
	{
	OpenFile("loc::m:\\dat\\test.dbf");       /*  This may change (obviously)  */
	GetFields();
	CleanUpAndQuit();
	}








