/* SOLMAIN.C */

// This file includes methods for:

// SOLAM (subclass of HWIMMAN)
// SOLWS (subclass of WSERV)
// SOLWIN (subclass of WIN)
// WAVEAO (subclass of ACTIVE)

#include <solipeg.g>
#include <solipeg.rsg>
#include <hwim.h>
#include <epoc.h>
#include <p_gen.h>

#define BORDER_WIDTH            6

GLREF_D PR_APPMAN * const       w_am ;  // Ptr to application manager
GLREF_D PR_WSERV *              w_ws ;  // Ptr to Window Server Client
GLREF_D WSERV_SPEC *            wserv_channel ; // Channel to WSERVER
GLREF_D TEXT *                  DatCommandPtr ; // Ptr to command line
GLREF_D TEXT *                  DatApp1 ; // Reserved application pointer

GLDEF_D PR_SOLENG *             soleng ; // Ptr to engine object

/************************************/
/* Solipeg Main Entry Point         */
/************************************/

GLDEF_C VOID main(VOID) {
    // Structures containing initializing data
    //  used in creating hwimman and wserv instances
    IN_HWIMMAN          app ;
    IN_WSERV            wserv ;

    // Dynamically link HWIM and OLIB categories
    p_linklib(0) ;

    // We'll be using our application resource file, the system resource
    //  file, the built-in cleanup functionality, and full S3a screen
    app.flags = FLG_APPMAN_RSCFILE |
                    FLG_APPMAN_SRSCFILE |
                    FLG_APPMAN_CLEAN |
                    FLG_APPMAN_FULLSCREEN ;

    // The category for all my classes is SOLIPEG
    wserv.com_cat = app.wserv_cat = p_getlibh(CAT_SOLIPEG_SOLIPEG) ;

    // My window server class is customized: C_SOLWS
    app.wserv_class = C_SOLWS ;

    // My command manager class is customized: C_SOLCM
    wserv.com_class = C_SOLCM ;

    // Create a new C_SOLAM object
    //  p_new() returns a ptr to the new solam object, which is
    //  sent the initialization method using app for initial
    //  appman/hwimman settings, and wserv for initial wserv settings.
    //  solam will create an instance of both solws and solcm according
    //  to these parameters
    p_send4(p_new(CAT_SOLIPEG_SOLIPEG, C_SOLAM), O_AM_INIT, &app, &wserv) ;
}


#pragma METHOD_CALL     // All the methods for WS/Window/AO follow


/******************************************/
/* Application Manager Subclass Method    */
/******************************************/

// Find the application's resource file
METHOD VOID solam_am_rscname(PR_SOLAM *self, TEXT *strName) {
    P_INFO      info ;

    // Build a suitable name string - possible machine dependent
    p_atos(strName, "\\APP\\SOLIPEG\\SOLI%02d.RZC", p_getlanguage()) ;

    // parse the name with respect to the executable's location
    f_fparse(strName, DatCommandPtr, strName, NULL) ;

    // if the file doesn't exist, call the superclass' original method
    if ( p_finfo(strName, &info) < 0 ) {
        p_supersend3(self, O_AM_RSCNAME, strName) ;
    }
}

/************************************/
/* Window Server Subclass Methods   */
/************************************/

// Initialize the window server client
METHOD VOID solws_ws_dyn_init(PR_SOLWS *self) {
    W_WINDATA       wd ;

    // This version will not run on S3 Classic, so pu up an error message
    if ( (wserv_channel->conn.info.version_id & WS_TYPE_MASK) != WS_TYPE_S3A ) {
        p_exit(hErrorDialog(E_GEN_NSUP, RS_S3AONLY)) ;
    }

    // Load the name of the environment variable used to store preferences
    hLoadResource(RS_ENV, &DatApp1) ;

    // Create engine - f_newsend() creates an object of the specified class in
    //  the specified category, sending it a specified message after creation,
    //  In this case, I tell the engine it's being created...
    soleng = f_newsend(CAT_SOLIPEG_SOLIPEG, C_SOLENG, O_ENG_INIT) ;

    // Create window to be the main client window connected to the server
    //  Here f_new() creates a C_SOLWIN object from the category SOLIPEG
    self->wserv.cli = f_new(CAT_SOLIPEG_SOLIPEG, C_SOLWIN) ;

    // Connect to window server with no backup bitmap
    wd.extent.tl.x = 0 ;
    wd.extent.tl.y = 0 ;
    wd.extent.width = wd.extent.height = wserv_channel->conn.info.pixels.y ;
    wd.background = W_WIN_BACK_NONE | W_WIN_BACK_GREY_NONE ;
    p_send5(    self->wserv.cli,
                O_WN_CONNECT,
                NULL,
                W_WIN_EXTENT | W_WIN_BACKGROUND,
                &wd) ;

    // Send the engine a NEW_GAME message
    p_send2(soleng, O_ENG_NEW_GAME) ;

    // Reposition the window according to preferences stored in the engine
    p_send3(self->wserv.cli, O_WN_REPOS, soleng->soleng.options.winpos) ;

    // Make the window visible
    p_send3(self->wserv.cli, O_WN_VISIBLE, WV_INITVIS) ;

    // Make a Status bar visible
    wsEnable() ;

    // Tell the engine to play a little sound
    p_send3(soleng, O_ENG_BEEP, RS_TITLE_SOUND) ;

    // Show initial help dialog
    self->wserv.help_index_id = RH_MAIN ;
    p_send3(self, O_WS_DO_HELP, RH_ABOUT) ;

    // Create a cursor, let the games begin
    p_send2(soleng, O_ENG_MAKE_CURSOR ) ;
}

/************************************/
/* Window Subclass Methods          */
/************************************/

// Reposition the window:
// uwPos = (0 == LEFT, 1 == CENTER, 2 == RIGHT)
METHOD VOID solwin_wn_repos(PR_SOLWIN *self, UWORD uwPos) {
    W_WINDATA       wd ;
    P_EXTENT        ext ;

    wInquireWindow(self->win.id, &wd) ;
    wInquireStatusWindow(W_STATUS_WINDOW_BIG, &ext) ;
    if ( uwPos == 0 ) {
        wd.extent.tl.x = 0 ;
    }
    else if ( uwPos == 1 ) {
        wd.extent.tl.x = (  wserv_channel->conn.info.pixels.x -
                            wd.extent.width ) / 2 ;
    }
    else {
        wd.extent.tl.x =    wserv_channel->conn.info.pixels.x -
                            wd.extent.width -
                            ext.width ;
    }
    wSetWindow( w_ws->wserv.cli->win.id,
                W_WIN_EXTENT,
                &wd ) ;
    p_send3(soleng, O_ENG_SETCURSOR, wd.extent.tl.x) ;
}

// REPLACEment for the WN_REDRAW method, which clears some rectangles
// as necessary
METHOD VOID solwin_wn_redraw(PR_SOLWIN *self, P_RECT *rectRedraw) {
    W_WINDATA   wd ;
    P_RECT      rect ;
    G_GC        gc ;

    gc.flags = G_GC_FLAG_BOTH_PLANES ;
    wBeginRedrawGC(self->win.id, rectRedraw, G_GC_MASK_GREY, &gc) ;

    wInquireWindow(self->win.id, &wd) ;

    // TOP
    rect.tl.x = rect.tl.y = 0 ;
    rect.br.x = wd.extent.width ;
    rect.br.y = BORDER_WIDTH ;
    if ( p_intrec(&rect, rectRedraw, &rect) ) {
        gClrRect(&rect, G_TRMODE_CLR) ;
    }

    // LHS
    rect.tl.x = rect.tl.y = 0 ;
    rect.br.x = BORDER_WIDTH ;
    rect.br.y = wd.extent.height ;
    if ( p_intrec(&rect, rectRedraw, &rect) ) {
        gClrRect(&rect, G_TRMODE_CLR) ;
    }

    // BOT
    rect.tl.x = 0 ;
    rect.tl.y = wd.extent.height - BORDER_WIDTH ;
    rect.br.x = wd.extent.width ;
    rect.br.y = wd.extent.height ;
    if ( p_intrec(&rect, rectRedraw, &rect) ) {
        gClrRect(&rect, G_TRMODE_CLR) ;
    }

    // RHS
    rect.tl.x = wd.extent.width - BORDER_WIDTH ;
    rect.tl.y = 0 ;
    rect.br.x = wd.extent.width ;
    rect.br.y = wd.extent.height ;
    if ( p_intrec(&rect, rectRedraw, &rect) ) {
        gClrRect(&rect, G_TRMODE_CLR) ;
    }

    p_send2(self, O_WN_DRAW) ;
    wEndRedraw() ;
}

// Pass the key stroke to the engine
METHOD VOID solwin_wn_key(PR_SOLWIN *self, INT keycode, INT mods) {
    // Tell the engine it's got a keypress
    p_send4(soleng, O_ENG_KEY, keycode, mods) ;
}

// Draw the window - border and board
METHOD VOID solwin_wn_draw(PR_SOLWIN *self) {
    gBorder2(W_BORDER_TYPE_1, W_BORD_CORNER_1 | W_BORD_SHADOW_ON) ;
    // Tell the engine to redraw the board
    p_send2(soleng, O_ENG_DRAW_BOARD) ;
}

/************************************/
/* Active Object Subclass Methods   */
/************************************/

// Simple REPLACEment for the AO_INIT method
METHOD VOID waveao_ao_init(PR_WAVEAO *self) {
    self->active.priority = PRIORITY_ACTIVE_VOICE ;
    // w_am is a ptr to the application manager
    p_send3(w_am, O_AM_ADD_TASK, self) ;
}

// Simple REPLACEment for the AO_CANCEL method
METHOD VOID waveao_ao_cancel(PR_WAVEAO *self) {
    p_playsoundcancel() ; // cancel any playing sound
    p_supersend2(self, O_AO_CANCEL) ; // let the regular method do its stuff
}

// ADDition of a WV_PLAY method
METHOD VOID waveao_wv_play( PR_WAVEAO * self,
                            TEXT *      name,
                            UINT        volume ) {
    if ( !self->active.isactive ) {
        p_playsounda(name, 0, volume, &self->active.stat) ; // start play
        self->active.isactive = TRUE ; // mark active
    }
}

